/*
  pro.cpp

  Copyright (c) 2023 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

#include "dmx.h"
#include "pro.h"

namespace ENTTEC::Pro {

void MsgProgramFlashPageRequest::iStream(std::shared_ptr<bufferstream> stream)
{
  if (stream->available() < sizeof(page))
    return stream->setstate(std::ios_base::failbit);

  stream->read(page, sizeof(page));

  if (!stream->available())
    stream->setstate(std::ios_base::eofbit);
};


void MsgProgramFlashPageRequest::oStream(std::shared_ptr<bufferstream> stream) const
{
  stream->write(page, sizeof(page));
};


void MsgProgramFlashPageReply::iStream(std::shared_ptr<bufferstream> stream)
{
  uint8_t buffer[sizeof(SUCCESS_OK)];
  success = false;

  if (stream->available() < sizeof(buffer))
      return stream->setstate(std::ios_base::failbit);

  stream->read(buffer, sizeof(buffer));

  if (!stream->available())
      stream->setstate(std::ios_base::eofbit);

  success = !std::memcmp(buffer, SUCCESS_OK, sizeof(buffer));
};


void MsgProgramFlashPageReply::oStream(std::shared_ptr<bufferstream> stream) const
{
  if (success)
    stream->write(reinterpret_cast<const uint8_t*>(SUCCESS_OK), sizeof(SUCCESS_OK));
  else
    stream->write(reinterpret_cast<const uint8_t*>(SUCCESS_FAIL), sizeof(SUCCESS_FAIL));
};


void MsgGetWidgetParametersRequest::iStream(std::shared_ptr<bufferstream> stream)
{
  *stream >> size;
};


void MsgGetWidgetParametersRequest::oStream(std::shared_ptr<bufferstream> stream) const
{
  *stream << size;
};


void MsgGetWidgetParametersReply::iStream(std::shared_ptr<bufferstream> stream)
{
  *stream >> version;
  *stream >> break_time;
  *stream >> mab_time;
  *stream >> rate;
  if (!stream->eof()) {
    std::vector<uint8_t> buffer;
    stream->read(buffer.data(), USER_CONFIGURATION_MAX);
    std::copy(buffer.begin(), buffer.end(), std::back_inserter(user_data));
  }
};


void MsgGetWidgetParametersReply::oStream(std::shared_ptr<bufferstream> stream) const
{
  *stream << fw_rev;
  *stream << fw_type;
  *stream << break_time;
  *stream << mab_time;
  *stream << rate;
  stream->write(user_data.data(), std::min(user_data.size(), USER_CONFIGURATION_MAX));
};


void MsgSetWidgetParametersRequest::iStream(std::shared_ptr<bufferstream> stream)
{
  uint16_t size;
  *stream >> size;
  *stream >> break_time;
  *stream >> mab_time;
  *stream >> rate;
  if (size)
  {
    std::vector<uint8_t> buffer;
    stream->read(buffer.data(), USER_CONFIGURATION_MAX);
    std::copy(buffer.begin(), buffer.end(), std::back_inserter(user_data));
  }
};


void MsgSetWidgetParametersRequest::oStream(std::shared_ptr<bufferstream> stream) const
{
  *stream << (uint16_t)std::min(user_data.size(), USER_CONFIGURATION_MAX);
  *stream << break_time;
  *stream << mab_time;
  *stream << rate;
  stream->write(user_data.data(), std::min(user_data.size(), USER_CONFIGURATION_MAX));
};


void MsgRecievedDmxPacket::iStream(std::shared_ptr<bufferstream> stream)
{
  *stream >> rx_errors;
  data.resize(stream->available());
  stream->read(data.data(), stream->available());
};


void MsgRecievedDmxPacket::oStream(std::shared_ptr<bufferstream> stream) const
{
  *stream << rx_errors;
  stream->write(data.data(), std::min(data.size(), DMX::E111_LAST_SLOT + 1));
};


void MsgOutputOnlySendDMX::iStream(std::shared_ptr<bufferstream> stream)
{
  stream->read(data.data(), DMX::E111_LAST_SLOT + 1);
};


void MsgOutputOnlySendDMX::oStream(std::shared_ptr<bufferstream> stream) const
{
  stream->write(data.data(), std::min(data.size(), DMX::E111_LAST_SLOT + 1));
  for (size_t i = data.size(); i <= DMX_LAST_SLOT_MIN; i++)
    *stream << (uint8_t)0;
};


void MsgSendRDMData::iStream(std::shared_ptr<bufferstream> stream)
{
  stream->read(data.data(), DMX::E111_LAST_SLOT + 1);
};


void MsgSendRDMData::oStream(std::shared_ptr<bufferstream> stream) const
{
  stream->write(data.data(), std::min(data.size(), DMX::E111_LAST_SLOT + 1));
};


void MsgRecieveDMXOnChange::iStream(std::shared_ptr<bufferstream> stream)
{
  uint8_t val;
  *stream >> val;
  if (val)
    mode = RxNotifyOnChange;
  else
    mode = RxNotifyAlways;
};


void MsgRecieveDMXOnChange::oStream(std::shared_ptr<bufferstream> stream) const
{
  if (mode == RxNotifyOnChange)
    *stream << (uint8_t)1;
  else
    *stream << (uint8_t)0;
};


void MsgRecievedDMXChanged::iStream(std::shared_ptr<bufferstream> stream)
{
  *stream >> start;
  stream->read(changed, sizeof(changed));
  stream->read(data.data(), 40);
};


void MsgRecievedDMXChanged::oStream(std::shared_ptr<bufferstream> stream) const
{
  *stream << start;
  stream->write(changed, sizeof(changed));
  stream->write(data.data(), std::min(data.size(), (size_t)40));
};


void MsgGetWidgetSerialReply::iStream(std::shared_ptr<bufferstream> stream)
{
  serial = BCDtoDecimal(stream->readType<uint32_t>());
};


void MsgGetWidgetSerialReply::oStream(std::shared_ptr<bufferstream> stream) const
{
  *stream << DecimalToBCD(serial);
};


void MsgSendRDMDiscovery::iStream(std::shared_ptr<bufferstream> stream)
{
  stream->read(request, sizeof(request));
};


void MsgSendRDMDiscovery::oStream(std::shared_ptr<bufferstream> stream) const
{
  stream->write(request, sizeof(request));
};

} // namespace ENTTEC::Pro
