/*
  universe.cpp

  Copyright (c) 2020 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

#include "universe.h"
#include <mutex>


namespace DMX {

/**
 * @brief Universe::Universe
 * @param timeout_period
 */
Universe::Universe(int timeout_period)
  : null_start_data({0})
  , device_class_(RECEIVER)
  , rx_timeout_period_(timeout_period)
  , status_(DMX_NULL)
  , active_data_slots_(1)  // start code
{
}


/**
 * @brief milliseconds since the last update
 * @return
 */
long Universe::age() const
{
  std::shared_lock lk_ctl(mtx_control);
  if (!last_updated_.time_since_epoch().count())
    return -1;  // universe has never been seen

  auto now = std::chrono::system_clock::now();
  auto elapsed = std::chrono::duration_cast<std::chrono::milliseconds>(now - last_updated_);

  return elapsed.count();
}


/**
 * @brief Universe::rxRate Rate (in Hz) at which the universe is being recv'd.
 * @return Hz
 *
 *  Calculated as the rolling mean of (the number of frames in
 *  the last rx_timeout_peroid ms) / rx_timeout_period seconds.
 */
double Universe::rxRate()
{
  rx_timeout_();
  std::shared_lock lk_ctl(mtx_control);
  //         updates     per      second
  return rx_times_.size() / (rx_timeout_period_ / 1000.0);
}


/**
 * @brief Universe::status
 * @return
 *
 * Like Schrödinger's cat, the status of the universe may exist in a superpostion
 * of states. Only by observing it will it collaps into a single status.
 */
uint8_t Universe::status() const
{
  std::shared_lock lk_ctl(mtx_control);
  return status_;
}


/**
 * @brief Universe::slot
 * @param address
 * @return
 */
uint8_t Universe::slot(const uint16_t address) const
{
  std::shared_lock lck(mtx_data);
  try {
    return null_start_data.at(address);
  } catch (...) {
    return 0;
  }
}


/**
 * @brief Universe::setStatus
 * @param val
 */
void Universe::setStatus(uint8_t val)
{
  if (val == RX_TIMEOUT)
    val = DMX_LOST;

  if (val == status_)
    return;

  {
    std::unique_lock lk_ctl(mtx_control);
    status_ = val;
  }
  do_callbacks_(cb_statusChange);
}


/**
 * @brief Universe::setValue
 * @param start
 * @param footprint
 * @param data
 */
void Universe::setValue(const uint16_t start, const uint16_t footprint,
                        const uint8_t * const data)
{
  if (!isEditable())
    return;

  // start and footprint valid?
  if (start < 1 || start + footprint > null_start_data.size())
    return;

  {
    std::unique_lock lk_data(mtx_data);
    std::copy_n(data, footprint, null_start_data.begin() + start);
  }
  {
    std::unique_lock lk_ctl(mtx_control);
    last_updated_ = std::chrono::system_clock::now();
  }

  // set active_data_slots to at least end of footprint
  if (start + footprint > active_data_slots_)
  {
    std::unique_lock lk_ctl(mtx_control);
    active_data_slots_ = start + footprint;
  }

  setStatus(DMX_ACTIVE);
  do_callbacks_(cb_dataChange);
}


/**
 * @brief Universe::setValue
 * @param address
 * @param value
 */
void Universe::setValue(const uint16_t address, const uint8_t value)
{
  setValue(address, 1, &value);
}


/**
 * @brief Universe::setData Accept new data from receiver
 * @param data varaibly sized data, beginning with a start code.
 *
 * The only setData method used to establish rx rate.
 */
void Universe::setData(const std::vector<uint8_t>& data)
{
  if (data.empty())
    return;                                 // no data
  switch (data.front()) {
  case E111_NULL_START:
    rx_dimmer_data_(data);
    break;
  default:
    if (rx_handlers_.count(data.front()))
      if (auto sp = rx_handlers_.at(data.front()).lock())
        (*sp)(data);                        // the owner is still holding the token
  }
}


/**
 * @brief Universe::onRxData
 * @param code
 * @param cb
 * @return
 */
std::shared_ptr<void> Universe::onRxData(uint8_t code,
                                         const std::function<void(const std::vector<uint8_t>&)> cb)
{
  std::unique_lock lk_ctl(mtx_control);
  // wrap the callback with a shared pointer
  auto sp = std::make_shared<std::function<void(const std::vector<uint8_t>&)>>(std::move(cb));
  // add callback to list (as a weak pointer)
  if (rx_handlers_.count(code))
    rx_handlers_[code] = sp;
  else
    rx_handlers_.insert({code, sp});
  // return token that caller must keep throughout it's scope
  return sp;
}


/**
 * @brief Universe::onData
 * @param cb
 * @return
 */
std::shared_ptr<void> Universe::onDataChange(const std::function<void(Universe*)> cb)
{
  std::unique_lock lk_ctl(mtx_control);
  // wrap the callback with a shared pointer
  auto sp = std::make_shared<std::function<void(Universe*)>>(std::move(cb));
  // add callback to list (as a weak pointer)
  cb_dataChange.push_back(sp);
  // return token that caller must keep throughout it's scope
  return sp;
}


/**
 * @brief Universe::onStatusChange
 * @param cb
 * @return
 */
std::shared_ptr<void> Universe::onStatusChange(const std::function<void(Universe*)> cb)
{
  std::unique_lock lk_ctl(mtx_control);
  // wrap the callback with a shared pointer
  auto sp = std::make_shared<std::function<void(Universe*)>>(std::move(cb));
  // add callback to list (as a weak pointer)
  cb_statusChange.push_back(sp);
  // return token that caller must keep throughout it's scope
  return sp;
}


/**
 * @brief Universe::rx_dimmer_data_
 * @param data
 */
void Universe::rx_dimmer_data_(const std::vector<uint8_t> &data)
{
  // accept variable lenth input, but never more than the limit
  const auto max_length = null_start_data.size();
  const auto length = data.size() < max_length ? data.size() : max_length;
  {
      std::unique_lock lk_data(mtx_data);
      if (length < max_length)
      null_start_data.fill(0);                                    // wipe old data
      std::copy_n(data.cbegin(), length, null_start_data.begin());  // copy new data
  }
  {
      std::unique_lock lk_ctl(mtx_control);
      active_data_slots_ = length;
  }
  rx_timeout_(true);                                              // update rx times
  do_callbacks_(cb_dataChange);                                   // run callbacks
}


/**
 * @brief Universe::rx_timeout_
 * @param add_now
 */
void Universe::rx_timeout_(bool add_now)
{
  auto now = std::chrono::system_clock::now();

  if (add_now)
    {
      setStatus(DMX_ACTIVE);
      std::unique_lock lk_ctl(mtx_control);
      last_updated_ = now;
      rx_times_.push(now);
    }
  else if (status_ == DMX_ACTIVE && age() >= E111_DATA_LOSS_TIMEOUT)
    // E1.11 DMX timeout period may be different the the rx_timeout_period
    setStatus(DMX_LOST);
  else if (age() >= rx_timeout_period_)
    // setStatus will be reimplimented to set the correct status
    setStatus(RX_TIMEOUT);

  auto elapsed = [] (auto& a, auto& b) {
      return std::chrono::duration_cast<std::chrono::milliseconds>(a - b).count();
    };

  long age;
  while (!rx_times_.empty())
  { // drop timestamps that are older than the timeout period
    age = elapsed(now, rx_times_.front());
    if (age < rx_timeout_period_)
      break;
    else
      {
        std::unique_lock lk_ctl(mtx_control);
        rx_times_.pop();
      }
  }
}


/**
 * @brief Universe::isEditable
 * @return
 */
bool Universe::isEditable() const
{
  std::shared_lock lk_ctl(mtx_control);
  return device_class_ == CONTROLLER;
}


/**
 * @brief Universe::activeSlots
 * @return
 */
uint16_t Universe::activeSlots() const
{
  std::shared_lock lk_ctl(mtx_control);
  return active_data_slots_;
}


/**
 * @brief Universe::doCallbacks
 * @param callbacks
 */
void Universe::do_callbacks_(std::vector<std::weak_ptr<const std::function<void(Universe*)>>> & callbacks)
{
  for (auto it = callbacks.cbegin(); it != callbacks.cend();)
    {
      if (auto sp = it->lock())
        { // the owner is still holding the token
          (*sp)(this);
          ++it;
        }
      else
        { // the owner has released the token
          std::unique_lock lk(mtx_control);
          it = callbacks.erase(it);
        }
    }
}

} // namespace DMX
