/*
  universe.h

  Copyright (c) 2020 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/
#pragma once

#include "dmx.h"

#include <chrono>
#include <cstdint>
#include <functional>
#include <map>
#include <memory>
#include <shared_mutex>
#include <queue>
#include <vector>

namespace DMX {

using DimmerData = std::array<uint8_t, E111_LAST_SLOT + 1>;  //!< Array of 513 bytes for basic data.

/**
 * @brief The Universe class
 *
 *  \cite DMX ANSI E1.11 describes may things about moving data over serial EIA-485-A
 *  links.  This class encapselates that data, regardless of how it was
 *  transmitted.
 */
class Universe {
  public:
    explicit Universe (int timeout_period = E111_DATA_LOSS_TIMEOUT);

    virtual long age() const;
    virtual double rxRate();
    virtual uint8_t status() const;
    virtual uint8_t slot(const uint16_t) const;

    virtual void setStatus(uint8_t);
    virtual void setValue(const uint16_t start, const uint16_t footprint,
                          const uint8_t* const data);
    virtual void setValue(const uint16_t address, const uint8_t value);
    virtual void setData(const std::vector<uint8_t> &);

    std::shared_ptr<void> onRxData(uint8_t, const std::function<void(const std::vector<uint8_t>&)>);
    std::shared_ptr<void> onDataChange(const std::function<void(Universe*)>);
    std::shared_ptr<void> onStatusChange(const std::function<void(Universe*)>);

    virtual bool isEditable() const;
    virtual uint16_t activeSlots() const;

    /// Known states of the universe Status value
    enum Status : uint8_t {
        DMX_NULL   = 0,  //!< uninitialized
        DMX_ACTIVE = 1,  //!< actively sending/receiving data
        DMX_LOST   = 2,  //!< no activity in E111_DATA_LOSS_TIMEOUT
        RX_TIMEOUT = 250 //!< no activity in the constructed timeout period
    };

  protected:
    inline void doDataCallbacks()   {do_callbacks_(cb_dataChange);}   //!< Data Change Callbacks
    inline void doStatusCallbacks() {do_callbacks_(cb_statusChange);} //!< Status Change Callbacks
    void do_callbacks_(std::vector<std::weak_ptr<const std::function<void(Universe*)>>>&);

    DimmerData null_start_data;                             //!< NULL Start Code data
    DeviceClass device_class_;                              //!< operating class of the universe
    mutable std::shared_mutex mtx_data;                     //!< memory protect Null Start data
    mutable std::shared_mutex mtx_control;                  //!< thread protected access control
    std::chrono::system_clock::time_point last_updated_;    //!< time of the latest update

  private:
    void rx_dimmer_data_(const std::vector<uint8_t> &);
    void rx_timeout_(bool add_now = false);

    const int rx_timeout_period_;
    uint8_t status_;                      //!< the operating state of the universe
    std::map<uint8_t, std::weak_ptr<const std::function<void(const std::vector<uint8_t>&)>>> rx_handlers_;
    std::queue<std::chrono::time_point<std::chrono::system_clock>> rx_times_;
    std::vector<std::weak_ptr<const std::function<void(Universe*)>>> cb_dataChange;
    std::vector<std::weak_ptr<const std::function<void(Universe*)>>> cb_statusChange;

  /**
   * @brief \cite DMX 8.7 Minimum number of data slots
   *
   * > There shall be no minimum number of data slots on the data link. DMX512 data packets
   * > with fewer than 512 slots may be transmitted, subject to the minimum timing requirements
   * > of this Standard – see clause 8.10 and figure 5.
   */
    uint16_t active_data_slots_;
};

} // namespace DMX
